import torch
import unittest
from qtorch.quant import *
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint


class TestQuant(unittest.TestCase):
    """
    random test on the behavior of nearest fixed point rounding
    """

    def test_fixed_random(self):
        S = lambda bits: 2 ** (bits)
        Q = lambda x, bits: torch.round(x * S(bits)) / S(bits)
        wl = 8
        quant = lambda x: fixed_point_quantize(x, wl=wl, fl=wl, clamp=False, rounding="nearest")

        N = int(1e8)
        for device in ["cpu", "cuda"]:
            x = torch.randn(N, device="cpu")
            oracle = Q(x, wl)
            target = quant(x)
            matched = torch.eq(oracle, target).all().item()
            self.assertTrue(matched)


if __name__ == "__main__":
    unittest.main()



import torch
import unittest
from qtorch.quant import *
from qtorch import FixedPoint, BlockFloatingPoint, FloatingPoint


class TestStochastic(unittest.TestCase):
    """
    invariant: quantized numbers cannot be greater than the maximum representable number
    or lower than the maximum representable number
    """

    def test_fixed(self):
        """test fixed point clamping"""
        for d in ["cpu", "cuda"]:
            for r in ["stochastic", "nearest"]:
                wl = 5
                fl = 4
                t_min = -(2 ** (wl - fl - 1))
                t_max = 2 ** (wl - fl - 1) - 2 ** (-fl)
                a = torch.linspace(-2, 2, steps=100, device=d)
                clamp_a = fixed_point_quantize(a, wl=wl, fl=fl, clamp=True, rounding=r)
                self.assertEqual(t_max, clamp_a.max().item())
                self.assertEqual(t_min, clamp_a.min().item())

                a = torch.linspace(-2, 2, steps=100, device=d)
                no_clamp_a = fixed_point_quantize(a, wl=wl, fl=fl, clamp=False, rounding=r)
                self.assertLess(t_max, no_clamp_a.max().item())
                self.assertGreater(t_min, no_clamp_a.min().item())

    def test_float(self):
        """test floating point clamping"""
        formats = [(6, 9), (5, 10), (5, 2)]

        for exp, man in formats:
            for d in ["cpu", "cuda"]:
                for r in ["stochastic", "nearest"]:
                    # test positive
                    a_max = 2 ** (2 ** (exp - 1)) * (1 - 2 ** (-man - 1))
                    a_min = 2 ** (-(2 ** (exp - 1)) + 1)
                    a = torch.Tensor([2 ** 50, a_min * 0.75, 2 ** (-50)]).to(device=d)
                    quant_a = float_quantize(a, exp=exp, man=man, rounding=r)
                    self.assertEqual(quant_a[0].item(), a_max)
                    self.assertAlmostEqual(quant_a[1].item(), a_min)
                    self.assertAlmostEqual(quant_a[2].item(), 0)

                    # test negative
                    a_max = -(2 ** (2 ** (exp - 1))) * (1 - 2 ** (-man - 1))
                    a_min = -(2 ** (-(2 ** (exp - 1)) + 1))
                    a = torch.Tensor([-(2 ** 50), a_min * 0.75, -(2 ** (-50))]).to(device=d)
                    quant_a = float_quantize(a, exp=exp, man=man, rounding=r)
                    self.assertEqual(quant_a[0].item(), a_max)
                    self.assertAlmostEqual(quant_a[1].item(), a_min)
                    self.assertAlmostEqual(quant_a[2].item(), 0)


if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn
from qtorch.quant import *
from collections import OrderedDict
import copy

__all__ = ["lower", "sequential_lower"]

SEQUENTIAL_LAYERS = [nn.Sequential, nn.ModuleList]  # TODO: Param List

DICT_LAYERS = [nn.ModuleDict]

CONV_LAYERS = [
    nn.Conv1d,
    nn.Conv2d,
    nn.Conv3d,
    nn.ConvTranspose1d,
    nn.ConvTranspose2d,
    nn.ConvTranspose3d,
    nn.Unfold,
    nn.Fold,
]

POOL_LAYERS = [
    nn.MaxPool1d,
    nn.MaxPool2d,
    nn.MaxPool3d,
    nn.MaxUnpool1d,
    nn.MaxUnpool2d,
    nn.MaxUnpool3d,
    nn.AvgPool1d,
    nn.AvgPool2d,
    nn.AvgPool3d,
    nn.FractionalMaxPool2d,
    nn.LPPool1d,
    nn.LPPool2d,
    nn.AdaptiveMaxPool1d,
    nn.AdaptiveMaxPool2d,
    nn.AdaptiveAvgPool2d,
    nn.AdaptiveMaxPool1d,
    nn.AdaptiveAvgPool1d,
    nn.AdaptiveMaxPool3d,
    nn.AdaptiveAvgPool3d,
]

PAD_LAYERS = [
    nn.ReflectionPad1d,
    nn.ReflectionPad2d,
    nn.ReplicationPad1d,
    nn.ReplicationPad2d,
    nn.ZeroPad2d,
    nn.ConstantPad1d,
    nn.ConstantPad2d,
    nn.ConstantPad3d,
]

ACTIVATION_LAYERS = [
    nn.ELU,
    nn.Hardshrink,
    nn.Hardtanh,
    nn.LeakyReLU,
    nn.LogSigmoid,
    nn.PReLU,
    nn.ReLU,
    nn.ReLU6,
    nn.RReLU,
    nn.SELU,
    nn.Sigmoid,
    nn.Softplus,
    nn.Softshrink,
    nn.Softsign,
    nn.Tanh,
    nn.Tanhshrink,
    nn.Threshold,
    nn.Softmin,
    nn.Softmax,
    nn.Softmax2d,
    nn.LogSoftmax,
]  # nn.AdaptiveLogSoftmaxWithLoss]

NORM_LAYERS = [
    nn.BatchNorm1d,
    nn.BatchNorm2d,
    nn.BatchNorm3d,
    nn.GroupNorm,
    nn.InstanceNorm1d,
    nn.InstanceNorm2d,
    nn.InstanceNorm3d,
    nn.LayerNorm,
    nn.LocalResponseNorm,
]

# Not supporting RNN layer

LINEAR_LAYERS = [nn.Linear, nn.Bilinear]

DROPOUT_LAYERS = [nn.Dropout, nn.Dropout2d, nn.Dropout3d, nn.AlphaDropout]

# Not supporting Sparse/Distance layers

LOSS_LAYERS = [
    nn.L1Loss,
    nn.MSELoss,
    nn.CrossEntropyLoss,
    nn.NLLLoss,
    nn.PoissonNLLLoss,
    nn.KLDivLoss,
    nn.BCELoss,
    nn.BCEWithLogitsLoss,
    nn.MarginRankingLoss,
    nn.HingeEmbeddingLoss,
    nn.MultiLabelMarginLoss,
    nn.SmoothL1Loss,
    nn.SoftMarginLoss,
    nn.MultiLabelSoftMarginLoss,  # nn.CosineEmbeddingLos,
    nn.MultiMarginLoss,
    nn.TripletMarginLoss,
]

LAYERS_TYPES = {
    "conv": CONV_LAYERS,
    "linear": LINEAR_LAYERS,
    "pool": POOL_LAYERS,
    "pad": PAD_LAYERS,
    "activation": ACTIVATION_LAYERS,
    "normalization": NORM_LAYERS,
    "dropout": DROPOUT_LAYERS,
    "loss": LOSS_LAYERS,
}


def _get_apply_lower_func(quant, layer_types=[]):
    def _insert_LP_layer(module):
        """Insert quant layer for all layers so long as in layer_types
        """
        lp_layer_types = []
        for layer_type in layer_types:
            assert layer_type in LAYERS_TYPES.keys()
            lp_layer_types += LAYERS_TYPES[layer_type]

        old_forward = module.forward
        if type(module) in lp_layer_types:
            module.forward = lambda *input: quant(old_forward(*input))
        else:
            return

    return _insert_LP_layer


def _get_return_sequential_lower_func(quant, layer_types=[]):
    def _insert_LP_layer(module):
        """Insert quant layer for all layers so long as in layer_types
        """
        if type(module) in SEQUENTIAL_LAYERS:
            for i, sub_module in enumerate(module.children()):
                module[i] = _insert_LP_layer(module[i])
            return module
        elif type(module) in DICT_LAYERS:
            for key, sub_module in module.items():
                module[key] = _insert_LP_layer(module[key])
            return module
        elif len(list(module.children())) != 0:
            for attribute_name in module.__dir__():
                attribute = getattr(module, attribute_name)
                if isinstance(attribute, nn.Module):
                    setattr(module, attribute_name, _insert_LP_layer(attribute))
            return module
        else:
            # Module has no submodule
            lp_layer_types = []
            for layer_type in layer_types:
                assert layer_type in LAYERS_TYPES.keys()
                lp_layer_types += LAYERS_TYPES[layer_type]

            if type(module) in lp_layer_types:
                module = nn.Sequential(module, quant)
            return module

    return _insert_LP_layer


def lower(
    model,
    layer_types=[],
    forward_number=None,
    backward_number=None,
    forward_rounding="stochastic",
    backward_rounding="stochastic",
):
    quant = Quantizer(
        forward_number, backward_number, forward_rounding, backward_rounding
    )
    lower_func = _get_apply_lower_func(quant, layer_types=layer_types)
    model.apply(lower_func)


def sequential_lower(
    model,
    layer_types=[],
    forward_number=None,
    backward_number=None,
    forward_rounding="stochastic",
    backward_rounding="stochastic",
):
    """Return a new model without touching the old one
    """
    quant = Quantizer(
        forward_number, backward_number, forward_rounding, backward_rounding
    )

    lower_func = _get_return_sequential_lower_func(quant, layer_types=layer_types)
    return lower_func(copy.deepcopy(model))


# def sequential_lower_(model,
#                       layer_types=[],
#                       forward_number=None, backward_number=None,
#                       forward_rounding="stochastic", backward_rounding="stochastic"):
#     """Modify the original model
#     """
#     quant = Quantizer(forward_number, backward_number,
#                       forward_rounding, backward_rounding)

#     lower_func = _get_return_sequential_lower_func(quant, layer_types=layer_types)
#     return _get_return_sequential_lower_func(quant, layer_types=layer_types)(model)



import torch
import torch.nn as nn
from .quant_function import *
import numpy as np

__all__ = ["Quantizer"]


class Quantizer(nn.Module):
    def __init__(
        self,
        forward_number=None,
        backward_number=None,
        forward_rounding="stochastic",
        backward_rounding="stochastic",
    ):
        super(Quantizer, self).__init__()
        self.quantize = quantizer(
            forward_number, backward_number, forward_rounding, backward_rounding
        )

    def forward(self, x):
        return self.quantize(x)



from setuptools import setup, find_packages

try:
    import torch

    has_dev_pytorch = "dev" in torch.__version__
except ImportError:
    has_dev_pytorch = False

# Base equirements
install_requires = [
    "torch>=1.5.0",
]
if has_dev_pytorch:  # Remove the PyTorch requirement
    install_requires = [
        install_require for install_require in install_requires if "torch" != re.split(r"(=|<|>)", install_require)[0]
    ]

setup(
    name="qtorch",
    version="0.2.0",
    description="Low-Precision Arithmetic Simulation in Pytorch",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Tianyi Zhang, Zhiqiu Lin, Guandao Yang, Christopher De Sa",
    author_email="tz58@cornell.edu",
    project_urls={
        "Documentation": "https://qpytorch.readthedocs.io",
        "Source": "https://github.com/Tiiiger/QPyTorch/graphs/contributors",
    },
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    python_requires=">=3.6",
    install_requires=install_requires,
)
